from __future__ import annotations

import logging
import re
from datetime import date, datetime
from numbers import Number

from soda.execution.data_type import DataType

logger = logging.getLogger(__name__)


class SqlDialect:
    """
    Extends ContractDataSource with all logic to builds the SQL queries.
    Specific ContractDataSource's can customize their SQL queries by subclassing SqlDialect,
    overriding methods of SqlDialect and returning the customized SqlDialect in ContractDataSource._create_sql_dialect()
    """

    def __init__(self):
        self.default_quote_char = self._get_default_quote_char()
        self.create_table_sql_type_dict: dict[str, str] = self._get_create_table_sql_type_dict()
        self.schema_check_sql_type_dict: dict[str, str] = self._get_schema_check_sql_type_dict()

    def _get_default_quote_char(self) -> str:
        return '"'

    def _get_create_table_sql_type_dict(self) -> dict[str, str]:
        return {
            DataType.TEXT: "VARCHAR(255)",
            DataType.INTEGER: "INT",
            DataType.DECIMAL: "FLOAT",
            DataType.DATE: "DATE",
            DataType.TIME: "TIME",
            DataType.TIMESTAMP: "TIMESTAMP",
            DataType.TIMESTAMP_TZ: "TIMESTAMPTZ",
            DataType.BOOLEAN: "BOOLEAN",
        }

    def _get_schema_check_sql_type_dict(self) -> dict[str, str]:
        return {
            DataType.TEXT: "character varying",
            DataType.INTEGER: "integer",
            DataType.DECIMAL: "double precision",
            DataType.DATE: "date",
            DataType.TIME: "time",
            DataType.TIMESTAMP: "timestamp without time zone",
            DataType.TIMESTAMP_TZ: "timestamp with time zone",
            DataType.BOOLEAN: "boolean",
        }

    def stmt_drop_schema_if_exists(self, database_name: str, schema_name: str) -> str:
        schema_name_quoted: str = self.quote_default(schema_name)
        return f"DROP SCHEMA IF EXISTS {schema_name_quoted} CASCADE"

    def stmt_create_schema_if_not_exists(self, database_name: str, schema_name: str) -> str:
        schema_name_quoted: str = self.quote_default(schema_name)
        return f"CREATE SCHEMA IF NOT EXISTS {schema_name_quoted} AUTHORIZATION CURRENT_USER"

    def stmt_select_table_names(
        self,
        database_name: str | None = None,
        schema_name: str | None = None,
        table_name_like_filter: str | None = None,
        include_table_name_like_filters: list[str] = None,
        exclude_table_name_like_filters: list[str] = None,
    ) -> str:
        """
        Builds the full SQL query to query table names from the data source metadata.
        """
        table_name_column: str = self.sql_information_schema_tables_table_column()
        # table_name_column_quoted: str = self.quote_default(table_name_column)
        information_schema_table_name_qualified: str = self.information_schema_table_name_qualified()

        sql = f"SELECT {table_name_column} \n" f"FROM {information_schema_table_name_qualified}"

        where_clauses: list[str] = self._filter_clauses_information_schema_table_name(
            database_name,
            schema_name,
            table_name_like_filter,
            include_table_name_like_filters,
            exclude_table_name_like_filters,
        )

        if where_clauses:
            where_clauses_sql = "\n  AND ".join(where_clauses)
            sql += f"\nWHERE {where_clauses_sql}"

        return sql

    def information_schema_table_name_qualified(self) -> str:
        information_schema_database_name: str = self.information_schema_database_name()
        information_schema_schema_name: str = self.information_schema_schema_name()
        information_schema_table_name: str = self.information_schema_table_name()
        return self.qualify_table(
            database_name=None, schema_name=information_schema_schema_name, table_name=information_schema_table_name
        )

    def information_schema_database_name(self) -> str | None:
        return None

    def information_schema_schema_name(self) -> str | None:
        return "information_schema"

    def information_schema_table_name(self) -> str:
        return "tables"

    def sql_information_schema_tables_database_column(self) -> str:
        """
        Name of the column in the information_schema table that contains the database name
        """
        return "table_catalog"

    def sql_information_schema_tables_schema_column(self) -> str:
        """
        Name of the column in the information_schema table that contains the schema name
        """
        return "table_schema"

    def sql_information_schema_tables_table_column(self) -> str:
        """
        Name of the column in the information_schema table that contains the table name
        """
        return "table_name"

    def _filter_clauses_information_schema_table_name(
        self,
        database_name: str | None = None,
        schema_name: str | None = None,
        table_name_like_filter: str | None = None,
        include_table_name_like_filters: list[str] = None,
        exclude_table_name_like_filters: list[str] = None,
    ) -> list[str]:
        """
        Builds the list of where clauses to query table names from the data source metadata.
        All comparisons are case-insensitive by converting to lower case.
        All column names are quoted.
        """

        where_clauses = []

        if database_name:
            database_column_name: str | None = self.sql_information_schema_tables_database_column()
            if database_column_name:
                # database_column_name: str = self.quote_default(database_column_name)
                database_name_lower: str = database_name.lower()
                where_clauses.append(f"LOWER({database_column_name}) = '{database_name_lower}'")

        if schema_name:
            schema_column_name: str | None = self.sql_information_schema_tables_schema_column()
            if schema_column_name:
                # schema_column_name: str = self.quote_default(schema_column_name)
                schema_name_lower: str = schema_name.lower()
                where_clauses.append(f"LOWER({schema_column_name}) = '{schema_name_lower}'")

        table_name_column = self.sql_information_schema_tables_table_column()
        # table_name_column = self.quote_default(table_name_column)

        if table_name_like_filter:
            where_clauses.append(f"LOWER({table_name_column}) LIKE '{table_name_like_filter.lower()}'")

        def build_table_matching_conditions(table_expressions: list[str], comparison_operator: str):
            conditions = []
            if table_expressions:
                for table_expression in table_expressions:
                    table_expression_lower: str = table_expression.lower()
                    conditions.append(f"LOWER({table_name_column}) {comparison_operator} '{table_expression_lower}'")
            return conditions

        if include_table_name_like_filters:
            sql_include_clauses = " OR ".join(build_table_matching_conditions(include_table_name_like_filters, "LIKE"))
            where_clauses.append(f"({sql_include_clauses})")

        if exclude_table_name_like_filters:
            sql_exclude_clauses = build_table_matching_conditions(exclude_table_name_like_filters, "NOT LIKE")
            where_clauses.extend(sql_exclude_clauses)

        return where_clauses

    def stmt_drop_test_table(self, database_name: str | None, schema_name: str | None, table_name: str) -> str:
        qualified_table_name = self.qualify_table(
            database_name=database_name, schema_name=schema_name, table_name=table_name
        )
        return f"DROP TABLE IF EXISTS {qualified_table_name}"

    def get_schema_check_sql_type(self, data_type: str) -> str:
        return self.schema_check_sql_type_dict.get(data_type)

    def get_schema_check_sql_type_text(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.TEXT)

    def get_schema_check_sql_type_time(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.TIME)

    def get_schema_check_sql_type_date(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.DATE)

    def get_schema_check_sql_type_timestamp(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.TIMESTAMP)

    def get_schema_check_sql_type_timestamp_tz(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.TIMESTAMP_TZ)

    def get_schema_check_sql_type_integer(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.INTEGER)

    def get_schema_check_sql_type_decimal(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.DECIMAL)

    def get_schema_check_sql_type_boolean(self) -> str:
        return self.schema_check_sql_type_dict.get(DataType.BOOLEAN)

    def qualify_table(
        self, database_name: str | None, schema_name: str | None, table_name: str, quote_table_name: bool = False
    ) -> str:
        """
        Creates a fully qualified table name, optionally quoting the table name
        """
        parts = [database_name, schema_name, self.quote_default(table_name) if quote_table_name else table_name]
        return ".".join([p for p in parts if p])

    def quote_default(self, identifier: str | None) -> str | None:
        return (
            f"{self.default_quote_char}{identifier}{self.default_quote_char}"
            if isinstance(identifier, str) and len(identifier) > 0
            else None
        )

    def default_casify(self, identifier: str) -> str:
        return identifier.lower()

    def literal(self, o: object) -> str:
        if o is None:
            return "NULL"
        elif isinstance(o, Number):
            return self.literal_number(o)
        elif isinstance(o, str):
            return self.literal_string(o)
        elif isinstance(o, datetime):
            return self.literal_datetime(o)
        elif isinstance(o, date):
            return self.literal_date(o)
        elif isinstance(o, list) or isinstance(o, set) or isinstance(o, tuple):
            return self.literal_list(o)
        elif isinstance(o, bool):
            return self.literal_boolean(o)
        raise RuntimeError(f"Cannot convert type {type(o)} to a SQL literal: {o}")

    def literal_number(self, value: Number):
        if value is None:
            return None
        return str(value)

    def literal_string(self, value: str):
        if value is None:
            return None
        return "'" + self.escape_string(value) + "'"

    def literal_list(self, l: list):
        if l is None:
            return None
        return "(" + (",".join([self.literal(e) for e in l])) + ")"

    def literal_date(self, date: date):
        date_string = date.strftime("%Y-%m-%d")
        return f"DATE '{date_string}'"

    def literal_datetime(self, datetime: datetime):
        return f"'{datetime.isoformat()}'"

    def literal_boolean(self, boolean: bool):
        return "TRUE" if boolean is True else "FALSE"

    def escape_string(self, value: str):
        return re.sub(r"(\\.)", r"\\\1", value)

    def escape_regex(self, value: str):
        return value



from __future__ import annotations

from soda.cloud.soda_cloud import SodaCloud as SodaCLSodaCloud
from soda.common.logs import Logs
from soda.scan import Scan


class CustomizedSodaClCloud(SodaCLSodaCloud):
    def __init__(
        self,
        host: str,
        api_key_id: str,
        api_key_secret: str,
        token: str | None,
        port: str | None,
        logs: Logs,
        scheme: str,
        default_data_source_properties: dict,
    ):
        super().__init__(host, api_key_id, api_key_secret, token, port, logs, scheme)
        self.default_data_source_properties: dict = default_data_source_properties

    def send_scan_results(self, scan: Scan):
        scan_results = self.build_scan_results(scan)
        scan_results["type"] = "sodaCoreInsertScanResults"
        scan_results["defaultDataSourceProperties"] = self.default_data_source_properties
        return self._execute_command(scan_results, command_name="send_scan_results")



from __future__ import annotations

import os

from soda.contracts.impl.yaml_helper import YamlFile, YamlHelper


class SodaCloud:
    def __init__(self, soda_cloud_file: YamlFile):
        logs = soda_cloud_file.logs
        configuration_dict = soda_cloud_file.get_dict() if soda_cloud_file.is_ok() else {}
        yaml_helper: YamlHelper = YamlHelper(logs=logs, yaml_file=soda_cloud_file)

        def get_configuration(key: str, default_value: str | None = None, is_required: bool = True) -> str | None:
            """
            Looks for the key in the configuration_dict, if it exists
            If not, in the corresponding environment variable
            If not applies the default value
            """
            environment_variable_name: str = f"SODA_CLOUD_{key.upper()}"
            default_value = os.environ.get(environment_variable_name, default_value)
            value = yaml_helper.read_string_opt(d=configuration_dict, key=key, default_value=default_value)
            if is_required and not isinstance(value, str):
                logs.error(f"Soda Cloud configuration '{key}' not provided as configuration nor environment variable")
            return value

        self.host: str = get_configuration(key="host", default_value="cloud.soda.io")
        self.api_key_id: str = get_configuration(key="api_key_id")
        self.api_key_secret: str = get_configuration(key="api_key_secret")
        self.token: str | None = get_configuration(key="token", is_required=False)
        self.port: str | None = get_configuration(key="port", is_required=False)
        self.scheme: str = get_configuration(key="scheme", is_required=False)



from __future__ import annotations

import dataclasses
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from numbers import Number
from typing import Dict

from soda.scan import Scan

from soda.contracts.impl.consistent_hash_builder import ConsistentHashBuilder
from soda.contracts.impl.logs import Location, Logs
from soda.contracts.impl.yaml_helper import QuotingSerializer, YamlFile, YamlHelper

logger = logging.getLogger(__name__)


class Check(ABC):

    def __init__(
        self,
        contract_file: YamlFile,
        data_source_name: str | None,
        database_name: str | None,
        schema_name: str | None,
        dataset_name: str,
        check_type: str,
        check_yaml: dict,
        logs: Logs,
    ):
        self.contract_file: YamlFile = contract_file
        self.data_source_name: str | None = data_source_name
        self.database_name: str | None = database_name
        self.schema_name: str | None = schema_name
        self.dataset_name: str = dataset_name
        self.type: str = check_type
        self.check_yaml: dict = check_yaml
        self.identity: str = self._create_identity()
        self.logs: Logs = logs
        self.skip: bool = False

    @abstractmethod
    def to_sodacl_check(self) -> str | dict | None:
        pass

    @abstractmethod
    def create_check_result(
        self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan
    ) -> CheckResult:
        pass

    @abstractmethod
    def _create_identity(self) -> str:
        pass


class CheckResult:

    def __init__(self, check: Check, outcome: CheckOutcome):
        self.check: Check = check
        self.outcome: CheckOutcome = outcome

    def __str__(self) -> str:
        return "\n".join(self.get_contract_result_str_lines())

    @abstractmethod
    def get_contract_result_str_lines(self) -> list[str]:
        """
        Provides the summary for the contract result logs, as well as the __str__ impl of this check result.
        Method implementations can use self._get_outcome_line(self)
        """

    def get_outcome_and_name_line(self) -> str:
        name_str: str = f" [{self.check.name}]" if self.check.name else ""
        return f"Check {self.get_outcome_str()}{name_str}"

    def get_outcome_str(self) -> str:
        if self.outcome == CheckOutcome.FAIL:
            return "FAILED"
        if self.outcome == CheckOutcome.PASS:
            return "passed"
        return "unverified"


class SchemaCheck(Check):

    def __init__(
        self,
        contract_file: YamlFile,
        data_source_name: str | None,
        database_name: str | None,
        schema_name: str | None,
        dataset_name: str,
        yaml_contract: dict,
        logs: Logs,
    ):
        super().__init__(
            contract_file=contract_file,
            data_source_name=data_source_name,
            database_name=database_name,
            schema_name=schema_name,
            dataset_name=dataset_name,
            check_type="schema",
            check_yaml=yaml_contract,
            logs=logs,
        )

        self.columns: dict[str, str] = {}
        self.optional_columns: list[str] = []

        yaml_helper = YamlHelper(logs=self.logs, yaml_file=self.contract_file)
        extra_columns: str | None = yaml_helper.read_string_opt(yaml_contract, "extra_columns")
        self.extra_columns_allowed: bool = "allowed" == extra_columns

        yaml_columns: list | None = yaml_helper.read_list(yaml_contract, "columns")
        if yaml_columns:
            for yaml_column in yaml_columns:
                column_name: str | None = yaml_helper.read_string(yaml_column, "name")
                data_type: str | None = yaml_helper.read_string_opt(yaml_column, "data_type")
                if column_name:
                    self.columns[column_name] = data_type

                is_column_optional = yaml_helper.read_bool_opt(yaml_column, "optional", default_value=False)
                if is_column_optional:
                    self.optional_columns.append(column_name)

    def _create_identity(self) -> str:
        return (
            ConsistentHashBuilder()
            .add_property("connection", self.data_source_name)
            .add_property("database", self.database_name)
            .add_property("schema", self.schema_name)
            .add_property("dataset", self.dataset_name)
            .add_property("type", self.type)
            .get_hash()
        )

    def to_sodacl_check(self) -> str | dict | None:
        column_names: dict[str, str | None] = {
            QuotingSerializer.quote(column_name): data_type for column_name, data_type in self.columns.items()
        }
        schema_fail_dict = {"when mismatching columns": column_names}
        if self.optional_columns:
            optional_column_names: list[str] = [
                QuotingSerializer.quote(column_name) for column_name in self.optional_columns
            ]
            schema_fail_dict["with optional columns"] = optional_column_names
        return {"schema": {"fail": schema_fail_dict}}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_measured_schema: list[dict] = scan_check_metrics_by_name.get("schema").get("value")
        measured_schema = {c.get("columnName"): c.get("sourceDataType") for c in scan_measured_schema}

        diagnostics = scan_check.get("diagnostics", {})

        columns_not_allowed_and_present: list[str] = diagnostics.get("present_column_names", [])
        columns_required_and_not_present: list[str] = diagnostics.get("missing_column_names", [])

        columns_having_wrong_type: list[DataTypeMismatch] = []
        scan_column_type_mismatches = diagnostics.get("column_type_mismatches", {})
        if scan_column_type_mismatches:
            for column_name, column_type_mismatch in scan_column_type_mismatches.items():
                expected_type = column_type_mismatch.get("expected_type")
                actual_type = column_type_mismatch.get("actual_type")
                columns_having_wrong_type.append(
                    DataTypeMismatch(column=column_name, expected_data_type=expected_type, actual_data_type=actual_type)
                )

        return SchemaCheckResult(
            check=self,
            outcome=CheckOutcome.from_scan_check(scan_check),
            measured_schema=measured_schema,
            columns_not_allowed_and_present=columns_not_allowed_and_present,
            columns_required_and_not_present=columns_required_and_not_present,
            columns_having_wrong_type=columns_having_wrong_type,
        )


class SchemaCheckResult(CheckResult):

    def __init__(
        self,
        check: Check,
        outcome: CheckOutcome,
        measured_schema: Dict[str, str],
        columns_not_allowed_and_present: list[str] | None,
        columns_required_and_not_present: list[str] | None,
        columns_having_wrong_type: list[DataTypeMismatch] | None,
    ):
        super().__init__(check, outcome)
        self.measured_schema: Dict[str, str] = measured_schema
        self.columns_not_allowed_and_present: list[str] | None = columns_not_allowed_and_present
        self.columns_required_and_not_present: list[str] | None = columns_required_and_not_present
        self.columns_having_wrong_type: list[DataTypeMismatch] | None = columns_having_wrong_type

    def get_contract_result_str_lines(self) -> list[str]:
        schema_check: SchemaCheck = self.check
        expected_schema: str = ",".join(
            [
                f"{c.get('name')}{c.get('optional')}{c.get('type')}"
                for c in [
                    {
                        "name": column_name,
                        "optional": "(optional)" if column_name in schema_check.optional_columns else "",
                        "type": f"={data_type}" if data_type else "",
                    }
                    for column_name, data_type in schema_check.columns.items()
                ]
            ]
        )

        lines: list[str] = [
            f"Schema check {self.get_outcome_str()}",
            f"  Expected schema: {expected_schema}",
            f"  Actual schema: {self.measured_schema}",
        ]
        lines.extend(
            [f"  Column '{column}' was present and not allowed" for column in self.columns_not_allowed_and_present]
        )
        lines.extend([f"  Column '{column}' was missing" for column in self.columns_required_and_not_present])
        lines.extend(
            [
                (
                    f"  Column '{data_type_mismatch.column}': Expected type '{data_type_mismatch.expected_data_type}', "
                    f"but was '{data_type_mismatch.actual_data_type}'"
                )
                for data_type_mismatch in self.columns_having_wrong_type
            ]
        )
        return lines


@dataclass
class CheckArgs:
    logs: Logs
    contract_file: YamlFile
    data_source_name: str | None
    database_name: str | None
    schema_name: str | None
    dataset_name: str
    filter: str | None
    check_type: str
    check_yaml: dict
    check_name: str | None
    check_name_was: str | None
    check_filter_sql: str | None
    threshold: Threshold
    location: Location
    yaml_helper: YamlHelper
    column: str | None = None
    missing_configurations: MissingConfigurations | None = None
    valid_configurations: ValidConfigurations | None = None


class CheckFactory(ABC):
    @abstractmethod
    def create_check(self, check_args: CheckArgs) -> Check | None:
        pass


class AbstractCheck(Check, ABC):

    threshold_keys = [
        "must_be_greater_than",
        "must_be_greater_than_or_equal_to",
        "must_be_less_than",
        "must_be_less_than_or_equal_to",
        "must_be",
        "must_not_be",
        "must_be_between",
        "must_be_not_between",
    ]

    validity_keys = [
        "invalid_values",
        "invalid_format",
        "invalid_regex_sql",
        "valid_values",
        "valid_format",
        "valid_regex_sql",
        "valid_min",
        "valid_max",
        "valid_length",
        "valid_min_length",
        "valid_max_length",
        "valid_values_reference_data",
    ]

    def __init__(self, check_args: CheckArgs):
        # name is initialized before super constructor because it's used in the _create_identity
        self.name: str | None = check_args.check_name
        # column is initialized before super constructor because it's used in the _create_identity
        self.column: str | None = check_args.column
        super().__init__(
            logs=check_args.logs,
            contract_file=check_args.contract_file,
            data_source_name=check_args.data_source_name,
            database_name=check_args.database_name,
            schema_name=check_args.schema_name,
            dataset_name=check_args.dataset_name,
            check_type=check_args.check_type,
            check_yaml=check_args.check_yaml,
        )
        self.name_was: str | None = check_args.check_name_was
        self.filter_sql: str | None = check_args.check_filter_sql
        self.missing_configurations: MissingConfigurations = check_args.missing_configurations
        self.valid_configurations: ValidConfigurations = check_args.valid_configurations
        self.threshold: Threshold = check_args.threshold
        self.location: Location = check_args.location

    def _create_identity(self) -> str:
        return self._create_identity_with_name(self.name)

    def _create_identity_with_name(self, name: str) -> str:
        return (
            ConsistentHashBuilder()
            .add_property("connection", self.data_source_name)
            .add_property("database", self.database_name)
            .add_property("schema", self.schema_name)
            .add_property("dataset", self.dataset_name)
            .add_property("type", self.type)
            .add_property("name", name)
            .get_hash()
        )

    def _create_sodacl_check_configs(self, check_specific_configs: dict | None = None) -> dict:
        check_configs: dict = {"identity": self.identity}
        if self.name:
            check_configs["name"] = self.name
        if self.name_was:
            identity_was: str = self._create_identity_with_name(self.name_was)
            check_configs["identity_was"] = identity_was
        if self.filter_sql:
            check_configs["filter"] = self.filter_sql
        if isinstance(check_specific_configs, dict):
            for key, value in check_specific_configs.items():
                if value is not None:
                    check_configs[key] = value
        return check_configs


class MissingCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type = check_args.check_type
        if check_type in ["no_missing_values", "missing_count", "missing_percent"]:
            threshold = check_args.threshold
            metric = check_type
            if check_type == "no_missing_values":
                metric = "missing_count"
                if threshold and not threshold.is_empty():
                    check_args.logs.error("Check type 'no_missing_values' does not allow for threshold keys must_...")
                else:
                    check_args.threshold = Threshold(equal=0)
            elif not threshold or threshold.is_empty():
                check_args.logs.error(f"Check type '{check_type}' requires threshold configuration")
            return MetricCheck(check_args=check_args, metric=metric)


class InvalidCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type = check_args.check_type
        if check_type not in ["no_invalid_values", "invalid_count", "invalid_percent"]:
            return None

        metric = "invalid_count" if check_type == "no_invalid_values" else check_type
        valid_configurations: ValidConfigurations = check_args.valid_configurations
        if valid_configurations and valid_configurations.valid_values_reference_data:
            return ReferenceDataCheck(check_args=check_args, metric=metric)

        threshold: Threshold | None = check_args.threshold
        if check_type == "no_invalid_values":
            if threshold and not threshold.is_empty():
                check_args.logs.error("Check type 'no_invalid_values' does not allow for threshold keys must_...")
            else:
                check_args.threshold = Threshold(equal=0)
        elif not threshold or threshold.is_empty():
            check_args.logs.error(f"Check type '{check_type}' requires threshold configuration")

        if not valid_configurations or not valid_configurations.has_non_reference_data_configs():
            check_args.logs.error(
                f"Check type '{check_type}' must have a validity configuration like {AbstractCheck.validity_keys}"
            )
        return MetricCheck(check_args=check_args, metric=metric)


class DuplicateCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type = check_args.check_type
        if check_type in ["no_duplicate_values", "duplicate_count", "duplicate_percent"]:
            threshold: Threshold | None = check_args.threshold
            metric = check_type
            if check_type == "no_duplicate_values":
                metric = "duplicate_count"
                if threshold and not threshold.is_empty():
                    check_args.logs.error("Check type 'no_duplicate_values' does not allow for threshold keys must_...")
                else:
                    check_args.threshold = Threshold(equal=0)
            elif not threshold or threshold.is_empty():
                check_args.logs.error(f"Check type '{check_type}' requires threshold configuration")

            return self.create_duplicate_check(check_args=check_args, metric=metric)

    def create_duplicate_check(self, check_args: CheckArgs, metric: str):
        return MetricCheck(check_args=check_args, metric=metric)


class SqlFunctionCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        metric: str = check_args.check_type
        return MetricCheck(check_args=check_args, metric=metric)


class RowCountCheckFactory(CheckFactory):

    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type: str = check_args.check_type
        if check_type in ["row_count", "rows_exist"]:
            threshold = check_args.threshold
            metric: str = check_type
            if check_type == "rows_exist":
                metric = "row_count"
                if not threshold.is_empty():
                    check_args.logs.error(
                        "Check type 'rows_exist' does not allow for threshold keys must_...",
                        location=check_args.location,
                    )
                check_args.threshold = Threshold(greater_than=0)
            elif threshold.is_empty():
                check_args.logs.error(
                    (
                        f"Check type '{check_type}' requires threshold configuration "
                        f"with keys like {AbstractCheck.threshold_keys}"
                    ),
                    location=check_args.location,
                )
            return MetricCheck(check_args=check_args, metric=metric)


class MetricCheck(AbstractCheck):

    def __init__(self, check_args: CheckArgs, metric: str):
        super().__init__(check_args)
        self.metric: str = metric

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_line = self.get_sodacl_check_line()
        sodacl_check_configs = self._create_sodacl_check_configs()

        if self.valid_configurations:
            sodacl_check_configs.update(self.valid_configurations.to_sodacl_check_configs_dict())
        if self.missing_configurations:
            sodacl_check_configs.update(self.missing_configurations.to_sodacl_check_configs_dict())

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        if "(" in self.metric:
            scan_metric_name = self.metric[: self.metric.index("(")]
            scan_metric_dict = scan_check_metrics_by_name.get(scan_metric_name, None)
        else:
            scan_metric_dict = scan_check_metrics_by_name.get(self.metric, None)
        metric_value: Number = scan_metric_dict.get("value") if scan_metric_dict else None
        return MetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=metric_value
        )

    def get_sodacl_check_line(self) -> str:
        sodacl_metric = self.get_sodacl_metric()
        sodacl_threshold: str = self.threshold.get_sodacl_threshold() if self.threshold else ""
        return f"{sodacl_metric} {sodacl_threshold}"

    def get_sodacl_metric(self) -> str:
        column_name: str = QuotingSerializer.quote(self.column)
        return f"{self.metric}({column_name})" if column_name else self.metric

    def get_sodacl_threshold(self) -> str:
        return self.threshold.get_sodacl_threshold() if self.threshold else "?"

    def get_metric_str(self) -> str:
        return self.get_sodacl_metric()

    def get_expected_str(self) -> str:
        return f"{self.get_metric_str()} {self.get_sodacl_threshold()}"


class MetricCheckResult(CheckResult):
    def __init__(
        self,
        check: Check,
        outcome: CheckOutcome,
        metric_value: Number,
    ):
        super().__init__(check, outcome)
        self.metric_value: Number = metric_value

    def get_contract_result_str_lines(self) -> list[str]:
        return [
            self.get_outcome_and_name_line(),
            f"  Expected {self.check.get_expected_str()}",
            f"  Actual {self.check.get_metric_str() } was {self.metric_value}",
        ]


class ReferenceDataCheck(MetricCheck):

    def __init__(self, metric: str, check_args: CheckArgs):
        super().__init__(check_args=check_args, metric=metric)
        self.valid_values_reference_data: ValidValuesReferenceData = (
            check_args.valid_configurations.valid_values_reference_data
        )

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = self._create_sodacl_check_configs()

        if self.valid_configurations:
            sodacl_check_configs.update(self.valid_configurations.to_sodacl_check_configs_dict())
        if self.missing_configurations:
            sodacl_check_configs.update(self.missing_configurations.to_sodacl_check_configs_dict())

        sodacl_check_line: str = (
            f"values in ({QuotingSerializer.quote(self.column)}) "
            f"must exist in {QuotingSerializer.quote(self.valid_values_reference_data.dataset)} "
            f"({QuotingSerializer.quote(self.valid_values_reference_data.column)})"
        )

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict = scan_check_metrics_by_name.get("reference", {})
        value: Number = scan_metric_dict.get("value")
        return MetricCheckResult(check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=value)


class UserDefinedMetricExpressionCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type: str = check_args.check_type
        if check_type == "metric_expression":
            return UserDefinedMetricExpressionCheck(check_args)


class UserDefinedMetricExpressionCheck(MetricCheck):
    def __init__(self, check_args: CheckArgs):
        check_yaml = check_args.check_yaml
        metric: str = check_args.yaml_helper.read_string_opt(check_yaml, "metric")
        super().__init__(check_args=check_args, metric=metric)
        self.expression_sql: str = check_yaml.get("expression_sql")

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = self._create_sodacl_check_configs({f"{self.metric} expression": self.expression_sql})

        sodacl_checkline_threshold = self.threshold.get_sodacl_threshold()
        sodacl_check_line = f"{self.get_sodacl_metric()} {sodacl_checkline_threshold}"

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict: dict = scan_check_metrics_by_name.get(self.metric, None)
        metric_value: Number = scan_metric_dict.get("value") if scan_metric_dict else None
        return MetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=metric_value
        )


class UserDefinedMetricQueryCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type: str = check_args.check_type
        if check_type == "metric_query":
            return UserDefinedMetricQueryCheck(check_args)


class UserDefinedMetricQueryCheck(MetricCheck):

    def __init__(self, check_args: CheckArgs):
        check_yaml = check_args.check_yaml
        metric: str = check_args.yaml_helper.read_string(check_yaml, "metric")
        super().__init__(check_args=check_args, metric=metric)
        self.query_sql: str = check_args.yaml_helper.read_string(check_yaml, "query_sql")

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = self._create_sodacl_check_configs({f"{self.metric} query": self.query_sql})

        sodacl_check_line: str = self.get_sodacl_check_line()

        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        scan_metric_dict: dict = scan_check_metrics_by_name.get(self.get_sodacl_check_line(), None)
        metric_value: Number = scan_metric_dict.get("value") if scan_metric_dict else None

        return MetricCheckResult(
            check=self, outcome=CheckOutcome.from_scan_check(scan_check), metric_value=metric_value
        )


class FreshnessCheckFactory(CheckFactory):
    def create_check(self, check_args: CheckArgs) -> Check | None:
        check_type = check_args.check_type
        if check_type.startswith("freshness_"):
            return FreshnessCheck(check_args)


class FreshnessCheck(AbstractCheck):

    def __init__(self, check_args: CheckArgs):
        super().__init__(check_args)

    def get_definition_line(self) -> str:
        column_name: str = QuotingSerializer.quote(self.column)
        return f"freshness({column_name}) {self.threshold.get_sodacl_threshold()}{self.get_sodacl_time_unit()}"

    def get_sodacl_time_unit(self) -> str:
        sodacl_time_unit_by_check_type = {
            "freshness_in_days": "d",
            "freshness_in_hours": "h",
            "freshness_in_minutes": "m",
        }
        return sodacl_time_unit_by_check_type.get(self.type)

    def to_sodacl_check(self) -> str | dict | None:
        sodacl_check_configs = self._create_sodacl_check_configs()
        sodacl_check_line: str = self.get_definition_line()
        return {sodacl_check_line: sodacl_check_configs}

    def create_check_result(self, scan_check: dict[str, dict], scan_check_metrics_by_name: dict[str, dict], scan: Scan):
        diagnostics: dict = scan_check["diagnostics"]
        freshness = diagnostics["freshness"]
        freshness_column_max_value = diagnostics["maxColumnTimestamp"]
        freshness_column_max_value_utc = diagnostics["maxColumnTimestampUtc"]
        now = diagnostics["nowTimestamp"]
        now_utc = diagnostics["nowTimestampUtc"]

        return FreshnessCheckResult(
            check=self,
            outcome=CheckOutcome.from_scan_check(scan_check),
            freshness=freshness,
            freshness_column_max_value=freshness_column_max_value,
            freshness_column_max_value_utc=freshness_column_max_value_utc,
            now=now,
            now_utc=now_utc,
        )


class FreshnessCheckResult(CheckResult):

    def __init__(
        self,
        check: Check,
        outcome: CheckOutcome,
        freshness: str,
        freshness_column_max_value: str,
        freshness_column_max_value_utc: str,
        now: str,
        now_utc: str,
    ):
        super().__init__(
            check=check,
            outcome=outcome,
        )
        self.freshness: str = freshness
        self.freshness_column_max_value: str = freshness_column_max_value
        self.freshness_column_max_value_utc: str = freshness_column_max_value_utc
        self.now: str = now
        self.now_utc: str = now_utc

    def get_contract_result_str_lines(self) -> list[str]:
        assert isinstance(self.check, FreshnessCheck)
        return [
            self.get_outcome_and_name_line(),
            f"  Expected {self.check.get_definition_line()}",
            f"  Actual freshness({self.check.column}) was {self.freshness}",
            f"  Max value in column was ...... {self.freshness_column_max_value}",
            f"  Max value in column in UTC was {self.freshness_column_max_value_utc}",
            f"  Now was ...................... {self.now}",
            f"  Now in UTC was ............... {self.now_utc}",
        ]


class MultiColumnDuplicateCheckFactory(DuplicateCheckFactory):

    def create_duplicate_check(self, check_args: CheckArgs, metric: str):
        columns: list[str] = check_args.yaml_helper.read_list_of_strings(check_args.check_yaml, "columns")
        return MultiColumnDuplicateCheck(check_args=check_args, metric=metric, columns=columns)


class MultiColumnDuplicateCheck(MetricCheck):

    def __init__(self, check_args: CheckArgs, metric: str, columns: list[str]):
        super().__init__(check_args=check_args, metric=metric)
        self.columns: list[str] = columns

    def get_sodacl_metric(self) -> str:
        # https://sodadata.slack.com/archives/C02J6Q493PY/p1714052722844239
        # column_str = (
        #     QuotingSerializer.quote(self.column) if self.column
        #     else ", ".join([QuotingSerializer.quote(column_name) for column_name in self.columns])
        # )
        column_str = self.column if self.column else ", ".join(self.columns)
        return f"{self.metric}({column_str})"


class CheckOutcome(Enum):
    PASS = "pass"
    FAIL = "fail"
    UNKNOWN = "unknown"

    @classmethod
    def from_scan_check(cls, scan_check: Dict[str, object]) -> CheckOutcome:
        scan_check_outcome = scan_check.get("outcome")
        if scan_check_outcome == "pass":
            return CheckOutcome.PASS
        elif scan_check_outcome == "fail":
            return CheckOutcome.FAIL
        return CheckOutcome.UNKNOWN


@dataclass
class DataTypeMismatch:
    column: str
    expected_data_type: str
    actual_data_type: str


def dataclass_object_to_sodacl_dict(dataclass_object: object) -> dict:
    def translate_to_sodacl_key(key: str) -> str:
        if "_sql" in key:
            key = key.replace("_sql", "")
        return key.replace("_", " ")

    dict_factory = lambda x: {translate_to_sodacl_key(k): v for (k, v) in x if v is not None}
    return dataclasses.asdict(dataclass_object, dict_factory=dict_factory)


@dataclass
class MissingConfigurations:
    missing_values: list[str] | list[Number] | None
    missing_regex_sql: str | None

    def to_sodacl_check_configs_dict(self) -> dict:
        return dataclass_object_to_sodacl_dict(self)


@dataclass
class ValidConfigurations:
    invalid_values: list[str] | list[Number] | None
    invalid_format: str | None
    invalid_regex_sql: str | None
    valid_values: list[str] | list[Number] | None
    valid_format: str | None
    valid_regex_sql: str | None
    valid_min: Number | None
    valid_max: Number | None
    valid_length: int | None
    valid_min_length: int | None
    valid_max_length: int | None
    valid_values_reference_data: ValidValuesReferenceData | None

    def to_sodacl_check_configs_dict(self) -> dict:
        sodacl_check_configs_dict = dataclass_object_to_sodacl_dict(self)
        sodacl_check_configs_dict.pop("valid values reference data", None)
        return sodacl_check_configs_dict

    def has_non_reference_data_configs(self) -> bool:
        return (
            self.invalid_values is not None
            or self.invalid_format is not None
            or self.invalid_regex_sql is not None
            or self.valid_values is not None
            or self.valid_format is not None
            or self.valid_regex_sql is not None
            or self.valid_min is not None
            or self.valid_max is not None
            or self.valid_length is not None
            or self.valid_min_length is not None
            or self.valid_max_length is not None
        )


@dataclass
class ValidValuesReferenceData:
    dataset: str
    column: str


@dataclass
class Threshold:
    """
    The threshold is exceeded when any of the member field conditions is True.
    To be interpreted as a check fails when the metric value is ...greater_than or ...less_than etc...
    """

    greater_than: Number | None = None
    greater_than_or_equal: Number | None = None
    less_than: Number | None = None
    less_than_or_equal: Number | None = None
    equal: Number | None = None
    not_equal: Number | None = None
    between: Range | None = None
    not_between: Range | None = None

    def get_sodacl_threshold(self) -> str:
        greater_bound: Number | None = (
            self.greater_than if self.greater_than is not None else self.greater_than_or_equal
        )
        less_bound: Number | None = self.less_than if self.less_than is not None else self.less_than_or_equal
        if isinstance(greater_bound, Number) and isinstance(less_bound, Number):
            if greater_bound > less_bound:
                return self.sodacl_threshold(
                    is_not_between=True,
                    lower_bound=less_bound,
                    lower_bound_included=self.less_than is not None,
                    upper_bound=greater_bound,
                    upper_bound_included=self.greater_than is not None,
                )
            else:
                return self.sodacl_threshold(
                    is_not_between=False,
                    lower_bound=greater_bound,
                    lower_bound_included=self.greater_than_or_equal is not None,
                    upper_bound=less_bound,
                    upper_bound_included=self.less_than_or_equal is not None,
                )
        elif isinstance(self.between, Range):
            return self.sodacl_threshold(
                is_not_between=False,
                lower_bound=self.between.lower_bound,
                lower_bound_included=True,
                upper_bound=self.between.upper_bound,
                upper_bound_included=True,
            )
        elif isinstance(self.not_between, Range):
            return self.sodacl_threshold(
                is_not_between=True,
                lower_bound=self.not_between.lower_bound,
                lower_bound_included=True,
                upper_bound=self.not_between.upper_bound,
                upper_bound_included=True,
            )
        elif self.greater_than is not None:
            return f"> {self.greater_than}"
        elif self.greater_than_or_equal is not None:
            return f">= {self.greater_than_or_equal}"
        elif self.less_than is not None:
            return f"< {self.less_than}"
        elif self.less_than_or_equal is not None:
            return f"<= {self.less_than_or_equal}"
        elif self.equal is not None:
            return f"= {self.equal}"
        elif self.not_equal is not None:
            return f"!= {self.not_equal}"

    @classmethod
    def sodacl_threshold(
        cls,
        is_not_between: bool,
        lower_bound: Number,
        lower_bound_included: bool,
        upper_bound: Number,
        upper_bound_included: bool,
    ) -> str:
        optional_not = "not " if is_not_between else ""
        lower_bound_bracket = "" if lower_bound_included else "("
        upper_bound_bracket = "" if upper_bound_included else ")"
        return f"{optional_not}between {lower_bound_bracket}{lower_bound} and {upper_bound}{upper_bound_bracket}"

    def is_empty(self) -> bool:
        return (
            self.greater_than is None
            and self.greater_than_or_equal is None
            and self.less_than is None
            and self.less_than_or_equal is None
            and self.equal is None
            and self.not_equal is None
            and self.between is None
            and self.not_between is None
        )


@dataclass
class Range:
    """
    Boundary values are inclusive
    """

    lower_bound: Number | None
    upper_bound: Number | None



from __future__ import annotations

import os
import re
from typing import Dict

from soda.contracts.impl.logs import Logs


class VariableResolver:

    def __init__(self, logs: Logs | None = None, variables: Dict[str, str] | None = None):
        # See also adr/03_exceptions_vs_error_logs.md
        self.logs: Logs = logs if logs else Logs()
        self.variables: Dict[str, str] | None = variables

    def resolve(self, text: str) -> str:
        return re.sub(
            pattern=r"\$\{([a-zA-Z_][a-zA-Z_0-9]*)\}",
            repl=lambda m: self._resolve_variable(m.group(1).strip()),
            string=text,
        )

    def _resolve_variable(self, variable_name: str) -> str:
        if self.variables is not None and variable_name in self.variables:
            return self.variables[variable_name]
        if variable_name in os.environ:
            return os.getenv(variable_name)
        self.logs.error(f"Variable '{variable_name}' not defined in the variables nor as environment variable")
        return ""



import logging

from soda.common.logs import Logs as SodaClLogs

from soda.contracts.impl.logs import Log, LogLevel, Logs

logger = logging.getLogger("soda.contracts.sodacl_log_converter")


class SodaClLogConverter(SodaClLogs):

    def __init__(self, logs: Logs):
        self.logs: Logs = logs

    def log_into_buffer(self, level, message, location, doc, exception):
        self.log(level=level, message=message, location=location, doc=doc, exception=exception)

    def log(self, level, message, location, doc, exception):
        log_level: LogLevel = LogLevel[level.name]
        self.logs._log(Log(log_level, message, location, exception))



from __future__ import annotations

import importlib
import logging
import textwrap
from abc import ABC, abstractmethod

from soda.execution.data_source import DataSource as SodaCLDataSource

from soda.contracts.impl.logs import Logs
from soda.contracts.impl.sodacl_log_converter import SodaClLogConverter
from soda.contracts.impl.sql_dialect import SqlDialect
from soda.contracts.impl.yaml_helper import YamlFile, YamlHelper

logger = logging.getLogger(__name__)


class ContractDataSource(ABC):
    """
    Takes configuration as input that is usually loaded from a data source YAML file.
    Is responsible for creating the connection in a with-block.  (see __enter__ and __exit__)
    Exposes methods to the rest of the engine that encapsulate interaction with the SQL connection.
    This class uses self.sql_dialect to build the SQL statement strings.
    """

    _KEY_TYPE = "type"
    _KEY_NAME = "name"
    _KEY_CONNECTION = "connection"
    _KEY_SPARK_SESSION = "spark_session"

    @classmethod
    def from_yaml_file(cls, data_source_yaml_file: YamlFile) -> ContractDataSource | None:
        data_source_yaml_dict = data_source_yaml_file.get_dict()
        if not isinstance(data_source_yaml_dict, dict):
            data_source_yaml_file.logs.error("No object content in data source YAML configuration")
            return None
        data_source_type: str = data_source_yaml_dict.get("type")
        if not isinstance(data_source_type, str):
            data_source_yaml_file.logs.error("No type property in data source YAML configuration")
            return None
        data_source_module_name: str = cls._create_data_source_module_name(data_source_type)
        data_source_class_name: str = cls._create_data_source_class_name(data_source_type)
        # dynamically load contracts data source
        try:
            module = importlib.import_module(f"soda.data_sources.{data_source_module_name}_contract_data_source")
            class_ = getattr(module, data_source_class_name)
            return class_(data_source_yaml_file)
        except ModuleNotFoundError as e:
            data_source_yaml_file.logs.error(
                f"Did you install the right package dependency for data source type {data_source_type}? \n"
                f"Could not instantiate contract data source {data_source_class_name} \n"
                f"{e}"
            )
            return None

    @classmethod
    def _create_data_source_module_name(cls, data_source_type: str) -> str:
        return data_source_type

    @classmethod
    def _create_data_source_class_name(cls, data_source_type: str) -> str:
        return f"{cls._camel_case_data_source_type(data_source_type)}ContractDataSource"

    @classmethod
    def _camel_case_data_source_type(cls, test_data_source_type) -> str:
        data_source_camel_case: dict[str, str] = {
            "bigquery": "BigQuery",
            "spark_df": "SparkDf",
            "sqlserver": "SQLServer",
            "mysql": "MySQL",
            "duckdb": "DuckDB",
        }
        if test_data_source_type in data_source_camel_case:
            return data_source_camel_case[test_data_source_type]
        else:
            return f"{test_data_source_type[0:1].upper()}{test_data_source_type[1:]}"

    def __init__(self, data_source_yaml_file: YamlFile):
        self.logs: Logs = data_source_yaml_file.logs

        # only initialized after the .open() method is called
        self.name: str | None = None
        # only initialized after the .open() method is called
        self.type: str | None = None

        self.data_source_yaml_file: YamlFile = data_source_yaml_file
        self.data_source_yaml_dict: dict = data_source_yaml_file.get_dict()

        yaml_helper: yaml_helper = YamlHelper(yaml_file=data_source_yaml_file, logs=self.logs)
        self.type = yaml_helper.read_string(self.data_source_yaml_dict, self._KEY_TYPE)
        self.name = yaml_helper.read_string(self.data_source_yaml_dict, self._KEY_NAME)

        self.sql_dialect: SqlDialect = self._create_sql_dialect()

        self.connection_yaml_dict: dict | None = yaml_helper.read_dict(self.data_source_yaml_dict, self._KEY_CONNECTION)

        # only initialized after the .open() method is called
        self.connection: object | None = None

        # Only if the open method has opened a connection should the close method close the connection
        # In the test suite this is used and later this may be used when passing in a user defined DBAPI connection
        self.close_connection_enabled: bool = False

        # The database name in the current connection context & current contract if applicable for this data source type
        # Updated in the ensure_connection method
        self.database_name: str | None = None
        # The schema name in the current connection context & current contract if applicable for this data source type
        # Updated in the ensure_connection method
        self.schema_name: str | None = None

    @abstractmethod
    def _create_sql_dialect(self) -> SqlDialect:
        raise NotImplementedError(f"{type(self).__name__} must override ContractDataSource._create_sql_dialect")

    def __str__(self) -> str:
        return self.name

    def open_connection(self) -> None:
        """
        Ensures that an open connection is available.
        After this method ends, the open connection must have the database_name and optionally
        the schema_name as context as far as these are applicable for the specific data source type.
        Potentially closes and re-opens the self.connection if the existing connection has a different database
        or schema context compared to the given database_name and schema_name for the next contract.
        """
        if self.connection is None:
            try:
                self.connection = self._create_connection(self.connection_yaml_dict)
                self.close_connection_enabled = True
            except Exception as e:
                self.logs.error(f"Could not connect to '{self.name}': {e}", exception=e)

    def set_contract_context(self, database_name: str | None, schema_name: str | None) -> None:
        """
        Ensures that an open connection is available.
        After this method ends, the open connection must have the database_name and optionally
        the schema_name as context as far as these are applicable for the specific data source type.
        Potentially closes and re-opens the self.connection if the existing connection has a different database
        or schema context compared to the given database_name and schema_name for the next contract.
        """
        self.database_name = database_name
        self.schema_name = schema_name

    @abstractmethod
    def _create_connection(self, connection_yaml_dict: dict) -> object:
        """
        self.connection_yaml_dict is provided as a parameter for convenience.
        Returns a new DBAPI connection based on the provided connection properties.
        The returned value will be saved in self.connection. See open() for details.
        Exceptions do not need to be handled.  They will be handled by the calling method.
        """

    def _log_connection_properties_excl_pwd(self, data_source_type: str, connection_yaml_dict: dict):
        dict_without_pwd = {
            k: v
            for k, v in connection_yaml_dict.items()
            if not "password" in k and not "pwd" in k and not "key" in k and not "secret" in k
        }
        logger.debug(f"{data_source_type} connection properties: {dict_without_pwd}")

    def _create_sodacl_data_source(
        self,
        sodacl_data_source_name: str,
    ) -> SodaCLDataSource:
        """
        Create SodaCL DataSource using the self.connection and sodacl_data_source_name
        """
        try:
            data_source_properties = self.connection_yaml_dict.copy()
            data_source_properties["database"] = self.database_name
            data_source_properties["schema"] = self.schema_name
            sodacl_data_source: SodaCLDataSource = SodaCLDataSource.create(
                logs=SodaClLogConverter(self.logs),
                data_source_name=sodacl_data_source_name,
                data_source_type=self.type,
                data_source_properties=data_source_properties,
            )
            sodacl_data_source.connection = self.connection
            return sodacl_data_source

        except Exception as e:
            self.logs.error(message=f"Could not create the data source: {e}", exception=e)

    def close_connection(self) -> None:
        """
        Closes te connection. This method will not throw any exceptions.
        Check errors with has_errors or assert_no_errors.
        """
        if self.close_connection_enabled and self.connection:
            try:
                self.connection.close()
            except Exception as e:
                logger.warning(f"Could not close the DBAPI connection: {e}")
            finally:
                self.connection = None

    def select_existing_test_table_names(self, database_name: str | None, schema_name: str | None) -> list[str]:
        sql = self.sql_dialect.stmt_select_table_names(
            database_name=database_name, schema_name=schema_name, table_name_like_filter="sodatest_%"
        )
        rows = self._execute_sql_fetch_all(sql)
        return [row[0] for row in rows]

    def _execute_sql_fetch_all(self, sql: str) -> list[tuple]:
        cursor = self.connection.cursor()
        try:
            sql_indented = textwrap.indent(text=sql, prefix="  #   ")
            logger.debug(f"SQL query fetchall: \n{sql_indented}")
            cursor.execute(sql)
            rows = cursor.fetchall()
            logger.debug(" | ".join([c[0] for c in cursor.description]))
            for row in rows:
                logger.debug(" | ".join([e for e in row]))
            return rows
        finally:
            cursor.close()

    def _execute_sql_update(self, sql: str) -> object:
        cursor = self.connection.cursor()
        try:
            sql_indented = textwrap.indent(text=sql, prefix="  # ")
            logger.debug(f"SQL update: \n{sql_indented}")
            updates = cursor.execute(sql)
            self.commit()
            return updates
        finally:
            cursor.close()

    def commit(self):
        self.connection.commit()

    def disable_close_connection(self) -> None:
        self.close_connection_enabled = False

    def enable_close_connection(self) -> None:
        self.close_connection_enabled = True

    def qualify_table(self, table_name: str) -> str:
        return self.sql_dialect.qualify_table(
            database_name=self.database_name, schema_name=self.schema_name, table_name=table_name
        )


class ClContractDataSource(ContractDataSource, ABC):

    def __init__(self, data_source_yaml_file: YamlFile):
        super().__init__(data_source_yaml_file=data_source_yaml_file)
        self.sodacl_data_source: SodaCLDataSource | None = None


class FileClContractDataSource(ClContractDataSource):

    def __init__(self, data_source_yaml_file: YamlFile):
        super().__init__(data_source_yaml_file=data_source_yaml_file)



from __future__ import annotations

import logging

from pyspark.sql import SparkSession
from soda.common import logs as soda_core_logs
from soda.data_sources.spark_df_contract_data_source import SparkDfContractDataSource
from soda.execution.data_source import DataSource as SodaCLDataSource
from soda.scan import Scan
from soda.scan import logger as scan_logger

from soda.contracts.contract import Contract, ContractResult
from soda.contracts.impl.contract_data_source import (
    ClContractDataSource,
    ContractDataSource,
)
from soda.contracts.impl.customized_sodacl_soda_cloud import CustomizedSodaClCloud
from soda.contracts.impl.logs import Location, Log, LogLevel, Logs
from soda.contracts.impl.plugin import Plugin
from soda.contracts.impl.soda_cloud import SodaCloud
from soda.contracts.impl.yaml_helper import YamlFile, YamlHelper

logger = logging.getLogger(__name__)


class ContractVerificationBuilder:

    def __init__(self):
        self.logs: Logs = Logs()
        self.data_source_yaml_file: YamlFile | None = None
        self.spark_session: object | None = None
        self.contract_files: list[YamlFile] = []
        self.soda_cloud_file: YamlFile | None = None
        self.plugin_files: list[YamlFile] = []
        self.variables: dict[str, str] = {}

    def with_contract_yaml_file(self, contract_yaml_file_path: str) -> ContractVerificationBuilder:
        if not isinstance(contract_yaml_file_path, str):
            self.logs.error(
                message=f"In ContractVerificationBuilder, parameter contract_yaml_file_path must be a string, but was {contract_yaml_file_path} ({type(contract_yaml_file_path)})"
            )
        self.contract_files.append(YamlFile(yaml_file_path=contract_yaml_file_path, logs=self.logs))
        return self

    def with_contract_yaml_str(self, contract_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(contract_yaml_str, str)
        self.contract_files.append(YamlFile(yaml_str=contract_yaml_str, logs=self.logs))
        return self

    def with_contract_yaml_dict(self, contract_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(contract_yaml_dict, dict)
        self.contract_files.append(YamlFile(yaml_dict=contract_yaml_dict, logs=self.logs))
        return self

    def with_data_source_yaml_file(self, data_source_yaml_file_path: str) -> ContractVerificationBuilder:
        assert isinstance(data_source_yaml_file_path, str)
        if self.data_source_yaml_file is not None:
            self.logs.error("Duplicate data source definition. Ignoring previous data sources.")
        self.data_source_yaml_file = YamlFile(yaml_file_path=data_source_yaml_file_path, logs=self.logs)
        return self

    def with_data_source_yaml_str(self, data_source_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(data_source_yaml_str, str)
        if self.data_source_yaml_file is not None:
            self.logs.error("Duplicate data source definition. Ignoring previous data sources.")
        self.data_source_yaml_file = YamlFile(logs=self.logs, yaml_str=data_source_yaml_str)
        return self

    def with_data_source_yaml_dict(self, data_source_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(data_source_yaml_dict, dict)
        if self.data_source_yaml_file is not None:
            self.logs.error("Duplicate data source definition. Ignoring previous data sources.")
        self.data_source_yaml_file = YamlFile(logs=self.logs, yaml_dict=data_source_yaml_dict)
        return self

    def with_data_source_spark_session(
        self, spark_session: object, data_source_yaml_dict: dict | None = None
    ) -> ContractVerificationBuilder:
        if data_source_yaml_dict is None:
            data_source_yaml_dict = {}
        assert isinstance(spark_session, object)
        assert isinstance(data_source_yaml_dict, dict)
        data_source_yaml_file = YamlFile(logs=self.logs, yaml_dict=data_source_yaml_dict)
        self.data_source_yaml_file = data_source_yaml_file
        self.spark_session = spark_session
        return self

    def with_soda_cloud_yaml_file(self, soda_cloud_yaml_file_path: str) -> ContractVerificationBuilder:
        assert isinstance(soda_cloud_yaml_file_path, str)
        if self.soda_cloud_file is not None:
            self.logs.error("Duplicate Soda Cloud definition. Ignoring previous data sources.")
        self.soda_cloud_file = YamlFile(yaml_file_path=soda_cloud_yaml_file_path, logs=self.logs)
        return self

    def with_soda_cloud_yaml_str(self, soda_cloud_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(soda_cloud_yaml_str, str)
        if self.soda_cloud_file is not None:
            self.logs.error("Duplicate Soda Cloud definition. Ignoring previous data sources.")
        self.soda_cloud_file = YamlFile(yaml_str=soda_cloud_yaml_str, logs=self.logs)
        return self

    def with_soda_cloud_yaml_dict(self, soda_cloud_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(soda_cloud_yaml_dict, dict)
        if self.soda_cloud_file is not None:
            self.logs.error("Duplicate Soda Cloud definition. Ignoring previous data sources.")
        self.soda_cloud_file = YamlFile(yaml_dict=soda_cloud_yaml_dict, logs=self.logs)
        return self

    def with_plugin_yaml_file(self, plugin_yaml_file_path: str) -> ContractVerificationBuilder:
        assert isinstance(plugin_yaml_file_path, str)
        self.plugin_files.append(YamlFile(yaml_file_path=plugin_yaml_file_path, logs=self.logs))
        return self

    def with_plugin_yaml_str(self, plugin_yaml_str: str) -> ContractVerificationBuilder:
        assert isinstance(plugin_yaml_str, str)
        self.plugin_files.append(YamlFile(yaml_str=plugin_yaml_str, logs=self.logs))
        return self

    def with_plugin_yaml_dict(self, plugin_yaml_dict: dict) -> ContractVerificationBuilder:
        assert isinstance(plugin_yaml_dict, dict)
        self.plugin_files.append(YamlFile(yaml_dict=plugin_yaml_dict, logs=self.logs))
        return self

    def with_variable(self, key: str, value: str) -> ContractVerificationBuilder:
        self.variables[key] = value
        return self

    def with_variables(self, variables: dict[str, str]) -> ContractVerificationBuilder:
        if isinstance(variables, dict):
            self.variables.update(variables)
        return self

    def build(self) -> ContractVerification:
        return ContractVerification(contract_verification_builder=self)

    def execute(self) -> ContractVerificationResult:
        contract_verification: ContractVerification = self.build()
        return contract_verification.execute()


class ContractVerification:

    @classmethod
    def builder(cls) -> ContractVerificationBuilder:
        return ContractVerificationBuilder()

    def __init__(self, contract_verification_builder: ContractVerificationBuilder):
        self.logs: Logs = contract_verification_builder.logs
        self.variables: dict[str, str] = contract_verification_builder.variables
        self.data_source: ContractDataSource | None = None
        self.contracts: list[Contract] = []
        self.soda_cloud: SodaCloud | None = None
        self.plugins: list[Plugin] = []
        self.contract_results: list[ContractResult] = []

        self._initialize_data_source(contract_verification_builder)
        self._initialize_soda_cloud(contract_verification_builder)
        self._initialize_contracts(contract_verification_builder)
        self._initialize_plugins(contract_verification_builder)

    def _initialize_data_source(self, contract_verification_builder: ContractVerificationBuilder) -> None:
        if isinstance(contract_verification_builder.data_source_yaml_file, YamlFile):
            data_source_yaml_file: YamlFile = contract_verification_builder.data_source_yaml_file
            if isinstance(data_source_yaml_file, YamlFile):
                data_source_yaml_file.parse(contract_verification_builder.variables)
                spark_session: SparkSession | None = contract_verification_builder.spark_session
                if spark_session is None:
                    data_source = ContractDataSource.from_yaml_file(data_source_yaml_file)
                else:
                    data_source = SparkDfContractDataSource(
                        data_source_yaml_file=data_source_yaml_file, spark_session=spark_session
                    )
                if isinstance(data_source, ContractDataSource):
                    self.data_source = data_source
                else:
                    self.logs.error(f"Error creating data source from {data_source_yaml_file}. See logs above.")

    def _initialize_contracts(self, contract_verification_builder: ContractVerificationBuilder) -> None:
        for contract_file in contract_verification_builder.contract_files:
            contract_file.parse(self.variables)
            if contract_file.is_ok():
                contract: Contract = Contract(
                    contract_file=contract_file,
                    logs=contract_file.logs,
                )
                self.contracts.append(contract)

    def _initialize_soda_cloud(self, contract_verification_builder: ContractVerificationBuilder) -> None:
        soda_cloud_file: YamlFile | None = contract_verification_builder.soda_cloud_file
        if isinstance(soda_cloud_file, YamlFile) and soda_cloud_file.exists():
            soda_cloud_file.parse(contract_verification_builder.variables)
            self.soda_cloud = SodaCloud(soda_cloud_file)

    def _initialize_plugins(self, contract_verification_builder) -> None:
        plugin_files_by_type: dict[str, list[YamlFile]] = {}
        for plugin_file in contract_verification_builder.plugin_files:
            plugin_file.parse(self.variables)
            if isinstance(plugin_file.dict, dict):
                plugin: str | None = plugin_file.dict.get("plugin")
                if isinstance(plugin, str):
                    plugin_files_by_type.setdefault(plugin, []).append(plugin_file)
                else:
                    self.logs.error(f"Key plugin is required in plugin YAML files")
            else:
                self.logs.error(f"Could not read plugin YAML file {plugin_file.file_path}")
        for plugin_name, plugin_yaml_files in plugin_files_by_type.items():
            for plugin_yaml_file in plugin_yaml_files:
                plugin_yaml_file.parse(self.variables)
            plugin: Plugin = Plugin.create(plugin_name, plugin_yaml_files, self.logs)
            if isinstance(plugin, Plugin):
                self.plugins.append(plugin)
            else:
                self.logs.error(f"Could not load plugin {plugin_name}")

    def __str__(self) -> str:
        return str(self.logs)

    def execute(self) -> ContractVerificationResult:
        contract_results: list[ContractResult] = []

        if isinstance(self.data_source, ContractDataSource):
            self.data_source.open_connection()
            try:
                if len(self.contracts) > 0:
                    for contract in self.contracts:
                        self.data_source.set_contract_context(
                            database_name=contract.database_name, schema_name=contract.schema_name
                        )
                        contract_result: ContractResult = self._verify(contract)
                        contract_results.append(contract_result)
                        for plugin in self.plugins:
                            plugin.process_contract_results(contract_result)
                else:
                    self.logs.error("No contracts specified")
            finally:
                self.data_source.close_connection()
        else:
            self.logs.error("No data source configured")

        return ContractVerificationResult(logs=self.logs, variables=self.variables, contract_results=contract_results)

    def _verify(self, contract: Contract) -> ContractResult:
        contract_data_source = self.data_source

        scan = Scan()

        scan_logs = soda_core_logs.Logs(logger=scan_logger)
        scan.set_verbose(True)

        sodacl_yaml_str: str | None = None
        try:
            sodacl_yaml_str = self._generate_sodacl_yaml_str(contract)
            logger.debug("Generated SodaCL:")
            logger.debug(sodacl_yaml_str)

            if not isinstance(contract_data_source, ClContractDataSource):
                raise NotImplementedError(f"Only ClDataSource's supported atm.  No support for {type(self).__name__}")

            if not isinstance(sodacl_yaml_str, str):
                self.logs.error("Bug: Empty SodaCL YAML string")
            else:
                scan._logs = scan_logs

                prefix_parts: list[str | None] = [contract.database_name, contract.schema_name]
                prefix_parts_str: list[str] = [
                    prefix_part for prefix_part in prefix_parts if isinstance(prefix_part, str)
                ]
                prefix_underscored: str = "_".join(prefix_parts_str)
                sodacl_data_source_name: str = f"{contract_data_source.name}_{prefix_underscored}"

                sodacl_data_source: SodaCLDataSource = contract_data_source._create_sodacl_data_source(
                    sodacl_data_source_name=sodacl_data_source_name
                )
                # noinspection PyProtectedMember
                scan._data_source_manager.data_sources[sodacl_data_source_name] = sodacl_data_source
                # Execute the contract SodaCL in a scan
                scan.set_data_source_name(sodacl_data_source_name)

                if self.soda_cloud:
                    scan_definition_name_parts: list[str] = [
                        sodacl_data_source_name,
                        contract.database_name,
                        contract.schema_name,
                        contract.dataset_name,
                    ]
                    scan_definition_name_parts_str: str = "/".join(
                        [part for part in scan_definition_name_parts if part is not None]
                    )
                    scan_definition_name = f"contract://{scan_definition_name_parts_str}"
                    scan.set_scan_definition_name(scan_definition_name)

                    default_data_source_properties = {
                        "type": contract_data_source.type,
                        "prefix": ".".join(prefix_parts_str),
                    }

                    # noinspection PyProtectedMember
                    scan._configuration.soda_cloud = CustomizedSodaClCloud(
                        host=self.soda_cloud.host,
                        api_key_id=self.soda_cloud.api_key_id,
                        api_key_secret=self.soda_cloud.api_key_secret,
                        token=self.soda_cloud.token,
                        port=self.soda_cloud.port,
                        logs=scan_logs,
                        scheme=self.soda_cloud.scheme,
                        default_data_source_properties=default_data_source_properties,
                    )

                if self.variables:
                    scan.add_variables(self.variables)

                scan.add_sodacl_yaml_str(sodacl_yaml_str)
                scan.execute()

        except Exception as e:
            self.logs.error(f"Data contract verification error: {e}", exception=e)

        # The scan warning and error logs are copied into self.logs and at the end of this
        # method, a SodaException is raised if there are error logs.
        self._append_scan_warning_and_error_logs(scan_logs=scan_logs, contract=contract)

        return ContractResult(
            data_source=contract_data_source,
            contract=contract,
            sodacl_yaml_str=sodacl_yaml_str,
            logs=self.logs,
            scan=scan,
        )

    def _generate_sodacl_yaml_str(self, contract: Contract) -> str:
        # Serialize the SodaCL YAML object to a YAML string
        sodacl_checks: list = []

        dataset_name: str = contract.dataset_name
        sodacl_yaml_object: dict = (
            {
                f"filter {dataset_name} [filter]": {"where": contract.filter_sql},
                f"checks for {dataset_name} [filter]": sodacl_checks,
            }
            if contract.filter_sql
            else {f"checks for {dataset_name}": sodacl_checks}
        )

        for check in contract.checks:
            if not check.skip:
                sodacl_check = check.to_sodacl_check()
                if sodacl_check is not None:
                    sodacl_checks.append(sodacl_check)
        yaml_helper: YamlHelper = YamlHelper(logs=self.logs)
        return yaml_helper.write_to_yaml_str(sodacl_yaml_object)

    def _append_scan_warning_and_error_logs(self, scan_logs: soda_core_logs.Logs, contract: Contract) -> None:
        level_map = {
            soda_core_logs.LogLevel.ERROR: LogLevel.ERROR,
            soda_core_logs.LogLevel.WARNING: LogLevel.WARNING,
            soda_core_logs.LogLevel.INFO: LogLevel.INFO,
            soda_core_logs.LogLevel.DEBUG: LogLevel.DEBUG,
        }
        for scan_log in scan_logs.logs:
            if scan_log.level in [soda_core_logs.LogLevel.ERROR, soda_core_logs.LogLevel.WARNING]:
                contracts_location: Location = (
                    Location(
                        file_path=contract.contract_file.get_file_description(),
                        line=scan_log.location.line,
                        column=scan_log.location.col,
                    )
                    if scan_log.location is not None
                    else None
                )
                contracts_level: LogLevel = level_map[scan_log.level]
                self.logs._log(
                    Log(
                        level=contracts_level,
                        message=f"SodaCL: {scan_log.message}",
                        location=contracts_location,
                        exception=scan_log.exception,
                    )
                )


class ContractVerificationResult:
    def __init__(self, logs: Logs, variables: dict[str, str], contract_results: list[ContractResult]):
        self.logs: Logs = logs
        self.variables: dict[str, str] = variables
        self.contract_results: list[ContractResult] = contract_results

    def failed(self) -> bool:
        """
        Returns True if there are execution errors or if there are check failures.
        """
        return not self.passed()

    def passed(self) -> bool:
        """
        Returns True if there are no execution errors and no check failures.
        """
        return not self.logs.has_errors() and all(contract_result.passed() for contract_result in self.contract_results)

    def has_errors(self) -> bool:
        return self.logs.has_errors()

    def has_failures(self) -> bool:
        return any(contract_result.failed() for contract_result in self.contract_results)

    def is_ok(self) -> bool:
        return not self.has_errors() and not self.has_failures()

    def assert_ok(self) -> ContractVerificationResult:
        errors_str: str | None = self.logs.get_errors_str() if self.logs.get_errors() else None
        if errors_str or any(contract_result.failed() for contract_result in self.contract_results):
            raise SodaException(message=errors_str, contract_verification_result=self)
        return self

    def __str__(self) -> str:
        blocks: list[str] = [str(self.logs)]
        for contract_result in self.contract_results:
            blocks.extend(self.__format_contract_results_with_heading(contract_result))
        return "\n".join(blocks)

    @classmethod
    def __format_contract_results_with_heading(cls, contract_result: ContractResult) -> list[str]:
        return [f"# Contract results for {contract_result.contract.dataset_name}", str(contract_result)]


class SodaException(Exception):
    """
    See also adr/03_exceptions_vs_error_logs.md
    """

    def __init__(
        self, message: str | None = None, contract_verification_result: ContractVerificationResult | None = None
    ):
        self.contract_verification_result: ContractVerificationResult | None = contract_verification_result
        message_parts: list[str] = []
        if message:
            message_parts.append(message)
        if self.contract_verification_result:
            message_parts.append(str(self.contract_verification_result))
        exception_message: str = "\n".join(message_parts)
        super().__init__(exception_message)
